@@ -26,14 +26,17 @@ if ($ARGV[0] eq "-l") {
 
 
 my ($file) = $ARGV [0] || die "must give a filename!!\n";
-
 my ($directory);
 my ($unit_name) = "";
 my ($definition_file) = $file;
+my ($has_get_type_subprogram) = 0;
+my ($enumerates) = "";   # Extra string to print for enumeration types
+my ($enumerates_clauses) = "";  # Representation clauses to add
+my ($signals) = ""; # List of all the signals for the widget
 
 ($directory, $file) = ($file =~ /^(.*\/)?([^\/]+)\.h$/);
 my ($hfile) = $file . ".h";
-my ($prefix) = ($file =~ /(g[dt]k)/);
+my ($prefix) = ($file =~ /(g[dt]k|gnome)/);
 substr ($prefix, 0, 1) = uc (substr ($prefix, 0, 1));
 
 if ($#ARGV >= 1) {
@@ -56,6 +59,7 @@ $file = uc ($file);
 $file = "GTKFILESELECTION" if ($file eq "GTKFILESEL");
 $file = "GTKFONTSELECTION" if ($file eq "GTKFONTSEL");
 $file = "GTK$1BUTTONBOX" if ($file =~ /GTK([VH]?)BBOX/);
+$file =~ s/-//g;  # Handle Gnome naming convention, e.g gnome-dock-band
 
 my (@cfile);
 
@@ -93,7 +97,7 @@ if ($list_mode) {
 		if ($tmp eq '_New') {
 		    $tmp = "Gtk_New";
 		}
-		$list{$tmp}++ unless ($tmp eq 'Get_Type');
+		$list{$tmp}++;
 		$from{$tmp} = $ARGV[0] . " $_";
 	    }
 	}
@@ -115,7 +119,6 @@ if ($list_mode) {
 			$from {$last} .= "  from misc.c";
 		    }
 		}
-		
 	    }
 	}
  	shift @ARGV;
@@ -134,6 +137,7 @@ close (FILE);
 
 
 my (%functions) = &parse_functions;
+
 my ($abstract) = 0;
 # 1 if the type is an abstract type
 # Set 1 on the previous line if you want to eventually create abstract types.
@@ -172,10 +176,10 @@ sub print_copyright_notice
 {
     print <<'EOF';
 -----------------------------------------------------------------------
---          GtkAda - Ada95 binding for the Gimp Toolkit              --
+--              GtkAda - Ada95 binding for Gtk+/Gnome                --
 --                                                                   --
---                     Copyright (C) 1998-1999                       --
---        Emmanuel Briot, Joel Brobecker and Arnaud Charlet          --
+--                     Copyright (C) 2001                            --
+--                         ACT-Europe                                --
 --                                                                   --
 -- This library is free software; you can redistribute it and/or     --
 -- modify it under the terms of the GNU General Public               --
@@ -202,6 +206,298 @@ sub print_copyright_notice
 EOF
 }
 
+######################################
+## Parse a list of parameters, and return
+##   Arg 1 = string: list of parameters as read in the C file
+##   Return= array that contains one element per parameter. See arg_to_ada
+##           for printing it to Ada.
+######################################
+
+sub parse_param_list {
+  my ($args) = shift;
+
+  # Associate pointers (return value) with the type instead of the
+  # name. Preserve comments indications (*/)
+  $args =~ s/(\w+)\s*\[\]/* $1/g;   # "*argv[]" arguments (see gnome-client.h)
+  $args =~ s/\s*\*(([^\/]|$))/* $1/g;
+
+  # Arrays are not fully supported (we would need to create a separate
+  # type. Instead, we just make sure that the file can be
+  # gnatchop-ed, and leave the responsability to the user for creating
+  # the type.
+  $args =~ s/(\w+)\s*(\w+)\s*\[(\d+)\]/Array_$1_$3 $2/g;
+
+  # Cleanup initial spaces and remove comments from the arguments list
+  $args =~ s/\/\*.*?\*\///g;
+  $args =~ s/\s+/ /g;
+  $args =~ s/(^|, ?)const /$1/g; # gnome-calculator.h
+
+  # Access-To-Subprograms are not supported, but at least we want to
+  # get some code that can be gnatchop-ed (gnome-procbar.h)
+  $args =~ s/, (\w+) \(\* *(\w+)\) *\([^\)]*\)/, Function_$1 $2/g;
+
+  return grep ($_ !~ /^\s*$/, split (/, ?/, $args));
+}
+
+#######################################
+##  Convert a C declaration for one parameter ("int a") to the equivalent
+##  Ada string
+##    Arg 1 = the string representing the C declaration
+##    Arg 2 = parameter number
+##    Arg 3 = conversion function to use
+##    return= ($name, $type)
+#######################################
+
+sub arg_to_ada {
+  $_ = shift;
+  my ($arg_num) = shift;
+  my ($convert) = shift;
+
+  return "" if ($_ eq "void");
+  s/\s+/ /g;
+  s/,//;
+  s/\);//;
+  my ($type, $name) = &parse_c_decl ($_, $arg_num);
+
+  $arg_num ++;
+
+  $type = &{$convert} ($type);
+
+  # Handling of widget parameters
+
+  if ($type =~ /\'Class/) {
+    $type =~ s/\'Class//;
+    if ($type =~ /$prefix\_$current_package/i) {
+      $type = "access $type\_Record";
+    }
+    else {
+      $type = "access $type\_Record'Class";
+    }
+
+    # Pointers arguments in C map to "out" parameters in Ada.
+    # However, pointers to pointers (char**) map to "out" parameters
+    # to the access type.
+  } elsif ($type =~ /(.*)\*$/) {
+    $type = $1;
+    if ($type !~ /^Gdk/) {
+      $type = "$1_Access" if ($type =~ /(.*)\*$/);
+      $type = "out $type";
+    }
+  }
+
+  return (&create_ada_name ($name), $type);
+}
+
+######################################
+## Parse the list of signals
+##   arg 1 = first line in the Class structure
+##   arg 2 = contents of the file (array)
+##   return= last line we reached while parsing the structure
+######################################
+
+sub parse_signals {
+  my ($line) = shift;
+  my (@deffile) = @_;
+  my ($return, $name, $params);
+
+  $signals = "   -------------\n"
+    . "   -- Signals --\n"
+    . "   -------------\n\n"
+    . "   --  <signals>\n"
+    . "   --  The following new signals are defined for this widget:\n"
+    . "   --\n";
+
+  while ($deffile[$line] !~ /};/) {
+    if ($deffile[$line] =~ /^\s*(\w+(\s*\*)?)\s*\(\*\s*(\w+)\)\s*\(([^\)]+)/)
+      {
+	($return, $name, $params) = ($1, $3, $4);
+	while ($deffile[$line] !~ /\);/) {
+	  $line++;
+	  $params .= $deffile[$line];
+	}
+
+	$params =~ s/\s+/ /g;
+	$params =~ s/\);\s*$//;
+	$signals .= "   --  - \"$name\"\n   --    ";
+	$signals .= ($return eq "void")? "procedure " : "function ";
+	$signals .=
+	  "Handler (Widget : access $prefix\_$current_package\_Record'Class";
+	$params =~ s/^[^,]+//;
+
+	if ($params ne "") {
+	  my ($arg_num) = 1;
+	  foreach (&parse_param_list ($params)) {
+	    my ($var, $type) = &arg_to_ada ($_, $arg_num, \&convert_ada_type);
+	    $arg_num++;
+	    $signals .= ";\n   --       $var : $type";
+	  }
+	}
+
+        $signals .= ")";
+	if ($return ne "void") {
+	  $signals .= "\n   --       "
+	    . "return " . &create_ada_name ($return);
+	}
+
+	$signals .= ";\n   --\n";
+      }
+    $line++;
+  }
+
+  $signals .= "   --  </signals>\n";
+
+  return $line;
+}
+
+######################################
+## Parse the enumeration type pointed to by the first parameters,
+##   arg 1 = first line in the Class structure
+##   arg 2 = contents of the file (array)
+##   return: line at which we stopped
+######################################
+
+sub parse_enums
+  {
+    my ($line) = shift;
+    my (@deffile) = @_;
+    my ($enum, $enum_name);
+
+    $enum = $deffile[$line];
+    while ($deffile[$line] !~ /;/) {
+      $line++;
+      $enum .= $deffile[$line];
+      chop $enum;  # so that multi-lines comments are correctly handled
+    }
+
+    # Remove comments
+    $enum =~ s/\/\*.*?\*\///g;
+    $enum =~ s/\s+/ /g;
+
+    ($enum, $enum_name) =
+      ($enum =~ /typedef enum {([^\}]*)}\s*([\w_]+)/);
+    $enum_name = &create_ada_name ($enum_name);
+    $enumerates .= "   type $enum_name is ";
+    $enum =~ s/\s//g;
+
+    # If there is a representation clause, use an Integer type instead
+    if ($enum =~ /<</) {
+      $enumerates .= "mod 2 ** 32;\n";
+      foreach (split (/,/, $enum)) {
+	my ($value, $num) = lc ($_);
+	$value =~ s/G[td]k_|Gnome_//i;
+	$value =~ s/^$current_package\_//i;
+	($value, $num) = ($value =~ /^([^=]+)=(.*)/);
+	$num = "2 ** $1" if ($num =~ /<<(\d*)\)?/);
+
+	$enumerates .= "   " . &create_ada_name ($value)
+	  . " : constant $enum_name := $num;\n";
+      }
+      $enumerates .= "\n";
+
+      # Else create an enumeration to match the C enum
+    } else {
+      $enumerates .= "(\n";
+      my ($first) = 1;
+      my ($has_representation) = 0;
+      foreach (split (/,/, $enum)) {
+	my ($value) = lc ($_);
+	$has_representation = 1 if ($value =~ s/=.*//);
+	$value =~ s/G[td]k_|Gnome_//i;
+	$value =~ s/^$current_package\_//i;
+	$enumerates .= ",\n" unless ($first);
+	$first = 0;
+	$enumerates .= "      " . &create_ada_name ($value);
+      }
+      $enumerates .= ");\n\n";
+
+      # Print the representation clause if necessary
+      if ($has_representation) {
+	$enumerates_clauses .= "   for $enum_name use (\n";
+	$first = 0;
+	foreach (split (/,/, $enum)) {
+	  my ($value, $repres) = lc ($_);
+	  $value =~ s/=([^,\)]*)//;
+	  $repres = $1 || $first;
+	  $value =~ s/G[td]k|Gnome_//i;
+	  $value =~ s/^$current_package\_//i;
+	  $enumerates_clauses .= ",\n" unless ($first == 0);
+	  $first = $value + 1;
+	  $enumerates_clauses .= "      " . &create_ada_name ($value)
+	    . " => $repres";
+	}
+	$enumerates_clauses .= ");\n\n";
+      }
+    }
+    return $line;
+  }
+
+######################################
+## Parse the definition of the widget, and check which fields should be
+## mapped to Ada
+##   arg 1 = first line in the Class structure
+##   arg 2 = contents of the file (array)
+##   return: line at which we stopped
+######################################
+
+sub parse_fields {
+  my ($line) = shift;
+  my (@deffile) = @_;
+
+  $line++ while ($deffile[$line] !~ /\{/);
+  $line++;
+  if ($deffile[$line] =~ /\/\*/) ## If we have a comment, skip it
+    {
+      while ($deffile[++$line] !~ /\*\//) {};
+      $line++;
+    }
+
+  ## Look for the parent widget.
+  ## This is the first non-blank line in the definition of the widget
+  ## structure. Note that this will be converted later on to Ada style,
+  ## we currently only store the C name
+
+  $line++ while ($deffile[$line] =~ /^\s*$/);
+  $parent = (split (/\s+/, $deffile[$line])) [1];
+
+  $line++;
+
+  ## Check all the fields (ie until the end of the structure)
+
+  while ($deffile[$line] !~ /\}/) {
+    my ($field);
+
+    # Search for the new field. Ignore empty lines and comments
+    while ($field =~ /^\s*$/) {
+      $field = $deffile[$line];
+
+      if ($field =~ /\/\*/) {
+	while ($deffile[$line] !~ /\*\//) {
+	  $line++;
+	  $field .= $deffile[$line];
+	}
+	$field =~ s$/\*.*\*/$$g;
+      }
+      $line++;
+    }
+
+    $field =~ s/\s+/ /g;
+    $field =~ s/ $//;
+    $field =~ s/^ //;
+    $field =~ s/\s\*/\* /g; ## attach the pointer to the type not to the field
+    $field =~ s/;//;
+    $field =~ s/:\s*\d+//;  ##  <type> <name> : <width>
+
+    # Types must be a single word, or split below will be confused
+    $field =~ s/unsigned ((g)?int)/$1/;
+    my ($type, $field) = split (/\s+/, $field);
+
+    print STDERR
+      "Create a function for the field $field (of type $type) [n]?";
+    my ($answer) = scalar (<STDIN>);
+    $fields {$field} = $type if ($answer =~ /^y/);
+  }
+  return $line;
+}
 
 ######################################
 ## Parse the definition file
@@ -213,80 +509,82 @@ sub parse_definition_file
   {
     my ($filename) = shift;
     local (*FILE);
-    my (@deffile);
+    my (@deffile, $line);
     open (FILE, $filename);
     @deffile = <FILE>;
     close (FILE);
-    
-    my ($line) = 0;
-    for ($line = 0; $line < $#deffile; $line ++)
-      {
-	last if ($deffile[$line] =~ /^STRUCT\s+_$file\s*$/i);
+
+    while ($line < $#deffile) {
+
+      # The Class record defines signals
+      if ($deffile[$line] =~ /^STRUCT\s+_${file}Class($|\s)/i) {
+	$line = &parse_signals ($line, @deffile);
       }
-    
-    if ($line < $#deffile) {
-      $deffile[$line] =~ /struct\s+_(.*)/;
-      $current_package = &create_ada_name ($1);
-      
-      my ($in_comment) = 0;
-      $line += 2;  ## skip the '{' line
-      if ($deffile[$line] =~ /\/\*/) ## If we have a comment, skip it
-	{
-	  1 while ($deffile[++$line] !~ /\*\//);
-	  $line++;
-	}
-      
-      ### Look for the parent widget
-      
-      $parent = (split (/\s+/, $deffile[$line])) [1];
-  
-      $line++;
-      while ($deffile[$line] !~ /\}/)
-	{
-	  if ($deffile[$line] =~ /\/\*/) ## If we have a comment, skip it
-	    {
-	      $line++ while ($deffile[$line] !~ /\*\//);
-	      $line++;
-	    }
-	  $line++ while ($deffile[$line] =~ /^\s*$/);
-	  chop ($deffile[$line]);
-	  $deffile[$line] =~ s/\s+/ /g;
-	  $deffile[$line] =~ s/ $//;
-	  $deffile[$line] =~ s/^ //;
-	  $deffile[$line] =~ s/ \*/\* /;  ## attach the pointer to the type not to the field
-	  $deffile[$line] =~ s/;//;
-	  my ($type, $field) = split (/ /, $deffile[$line]);
-	  
-	  print STDERR "Create a function for the field $field (of type $type) [n]?";
-	  my ($answer) = scalar (<STDIN>);
-	  if ($answer =~ /^y/)
-	    {
-	      $fields {$field} = $type;
-	    }
-	  $line ++;
-	}
+
+      # Enumeration types should also be visible in the Ada specs
+      elsif ($deffile[$line] =~ /^typedef enum/) {
+	$line = &parse_enums ($line, @deffile);
+      }
+
+      # The widget record defines all the fields
+      elsif ($deffile[$line] =~ /^struct\s+_($file)($|\s)/i) {
+
+	# Get the package name from the name of the structure. Given the
+	# various casing used in Gtk+/Gnome, there are a few exceptions that
+	# we need to handle manually
+
+	$current_package = $1;
+	$current_package = "Dentry_Edit" if ($1 eq "GnomeDEntryEdit");
+	$current_package = "Ui_Info"     if ($1 eq "GnomeUIInfo");
+	$current_package = "Mdi"         if ($1 eq "GnomeMDI");
+	$current_package = "Mdi_$1"      if ($1 =~ /GnomeMDI(.+)/);
+	$current_package = &create_ada_name ($current_package);
+
+	$line = &parse_fields ($line, @deffile);
+      }
+
+      $line ++;
     }
   }
-    
+
+###################################
+##  Sorting function used to generate the subprograms in a correct order
+###################################
+
+sub func_sort () {
+  # "_newv" must be supported as well (gnome-canvas.h)
+  return (-1) if ($a =~ /_newv?$/ || $a =~ /_newv?_/);
+  return (1)  if ($b =~ /_newv?$/ || $b =~ /_newv?_/);
+
+  return &ada_func_name ($a) cmp &ada_func_name ($b);
+}
+
 ###################################
 ##  Generates the package specification file
 ###################################
+
 sub generate_specifications
   {
-    %with_list = ();
+    my ($parent_prefix);
+
+    %with_list = ("with Gtk" => 1);
     @output = ();
     $with_list {"with " . &package_name ($parent)} ++;
-    foreach (sort {&ada_func_name ($a) cmp &ada_func_name ($b)} keys %functions)
+    foreach (sort func_sort keys %functions)
       {
 	&print_declaration ($_, @{$functions{$_}});
       }
-    my ($parent_string) = &package_name ($parent). ".$prefix\_" . &create_ada_name ($parent);
-    if ($parent eq "Root_Type") {
-	$parent_string = "Root_Type";
-    }
-    else {
-	$parent_string = "$parent_string\_Record";
-    }
+
+    $parent_prefix = "Gtk"   if ($parent =~ /^gtk/i);
+    $parent_prefix = "Gdk"   if ($parent =~ /^gdk/i);
+    $parent_prefix = "Gnome" if ($parent =~ /^gnome/i);
+
+    my ($parent_string) = &package_name ($parent). ".$parent_prefix\_"
+	. &create_ada_name ($parent);
+    $parent_string = ($parent eq "Root_Type") ?
+	"Root_Type" : "$parent_string\_Record";
+
+    unshift (@output, $enumerates);
 
     unshift (@output, "   type $prefix\_$current_package is access all "
 	     . "$prefix\_$current_package\_Record'Class;\n\n");
@@ -294,16 +592,27 @@ sub generate_specifications
 	     . ($abstract ? "abstract " : "")
 	     . "new " . $parent_string . " with private;\n");
     unshift (@output, "package $prefix.$current_package is\n\n");
-    
+
+    push (@output, $signals);
     push (@output, "\nprivate\n");
     push (@output, "   type $prefix\_$current_package\_Record is "
 	  . ($abstract ? "abstract " : "")
-	  . "new ".
-	  &package_name ($parent). ".$prefix\_", &create_ada_name ($parent).
-	  "_Record with null record;\n\n");
-    
+	  . "new ",
+	  $parent_string = ($parent eq "Root_Type") ?
+	      "Root_Type" : "$parent_string",
+	  " with null record;\n\n");
+
+    push (@output, $enumerates_clauses);
+
+    if ($has_get_type_subprogram == -1) {
+      push (@output,
+	    "   pragma Import (C, Get_Type, \""
+	    . lc ("$prefix\_$current_package\_get_type") . "\");"
+	    . "\n");
+    }
+
     push (@output, "end $prefix.$current_package;\n");
-    
+
     print "\n", join (";\n", sort keys %with_list), ";\n"; 
     print "\n", join ("", @output);
   }
@@ -313,16 +622,17 @@ sub generate_specifications
 ###################################
 sub generate_body
   {
-    %with_list = ("with Gdk; use Gdk" => 1,
+    %with_list = ("with Gtk; use Gtk" => 1,
+                  "with Gdk; use Gdk" => 1,
 		  "with System" => 1);
     @output = ();
     push (@output, "package body $prefix.$current_package is\n\n");
-    
-    foreach (sort {&ada_func_name ($a) cmp &ada_func_name ($b)} keys %functions)
+
+    foreach (sort func_sort keys %functions)
       {
 	&print_body ($_, @{$functions{$_}});
       }
-    
+
     push (@output, "end $prefix.$current_package;\n");
 
     ## If there is indeed a body
@@ -365,10 +675,13 @@ sub create_ada_name
     my ($i);
     my ($char);
 
-    $entity =~ s/([^_])([A-Z])/$1_$2/g;
-    
+    $entity =~ s/-/_/g;
+
+    # Put an underscore before each upper-case letter
+    $entity =~ s/([^_])([A-Z0-9])/$1_$2/g;
+
     substr ($entity, 0, 1) = uc (substr ($entity, 0, 1));
-    
+
     for ($i = 1; $i < length ($entity); $i ++)
       {
 	$char = substr ($entity, $i, 1); 
@@ -377,11 +690,16 @@ sub create_ada_name
 	    substr ($entity, $i + 1, 1) = uc (substr ($entity, $i + 1, 1));
 	  }
       }
-    $entity =~ s/^G[dt]k_?//;
+    $entity =~ s/^(G[dt]k|Gnome)_?//;
 
     return "The_Type" if ($entity eq "Type");
     return "The_End" if ($entity eq "End");
-    return "The_Digits" if ($entity eq "Digits");    
+    return "The_Digits" if ($entity eq "Digits");
+    return "GRange" if ($entity eq "Range");     # gtk-range.h
+    return "GEntry" if ($entity eq "Entry");     # gnome-entry.h
+    return "Accepted" if ($entity eq "Accept");  # gnome-icon-item.h
+    return "Modifier" if ($entity eq "Mod");     # gnome-stock.h
+    return "Sub_Type" if ($entity eq "Subtype"); # gnome-stock.h
     return $entity;
   }
 
@@ -392,11 +710,9 @@ sub create_ada_name
 sub package_name
   {
     my ($entity) = shift;
-    
-    $entity =~ s/(G[td]k)/$1./;
+
+    $entity =~ s/(G[dt]k|Gnome)/$1./;
     $entity =~ s/([a-z])([A-Z])/$1_$2/g;
-    $entity =~ s/Gtk\.Range/Gtk\.Gtk_Range/;
-    $entity =~ s/Gtk\.Entry/Gtk\.Gtk_Entry/;    
     return $entity;
   }
 
@@ -407,44 +723,52 @@ sub package_name
 #######################
 sub parse_functions
   {
-    my ($func_name);
-    my ($return);
+    my ($func_name, $return);
     my (@arguments);
     my (%functions);
-    
+
     while ($_ = shift @cfile)
       {
-	chop;
-	s/ \*/\* /g;
-	if (/^(\S+)\s+(g[dt]k_\S+)\s*\((.*)/)
-	  {
-	    $func_name = $2;
-	    $return = $1;
-	    my ($tmp) = $3;
-	    $tmp =~ s/,\s*$//;
-	    @arguments = split (/,/, $tmp);
-	    if ($3 !~ /\);/)
-	      {
-		while ($_ = shift @cfile)
-		  {
-		    chop;
-		    /\s+(.*)/;
-		    push (@arguments, grep ($_ !~ /^$/, split (/,/, $1)));
-		    last if ($1 =~ /\);/);
-		  }
-	      }
-	    if ($unit_name eq ""
-	       || $func_name =~ /$prefix\_$unit_name/i) {
-	      push (@{$functions{$func_name}},
-		    $return, @arguments);
-	    }
+	# If this looks like a subprogram start (<return_type>  <name>)
+
+	if (/^(\w\S+(\s+\*)?)\s*((g[dt]k|gnome)_\S+)/) {
+	  chop;
+
+	  # Read the whole subprogram definition at once. This makes it
+	  # easier to ignore comments, to handle subprograms defined on
+	  # multiple lines, ...
+
+	  while ($_ !~ /\);/) {
+	    $_ .= shift @cfile;
+	    chop;
 	  }
+
+	  /^(\w\S+(\s+\*)?)\s*(\S+)\s*\((.*)\);/;
+	  $return = (&parse_param_list ($1)) [0];
+	  $func_name = $3;
+	  @arguments = &parse_param_list ($4);
+
+	  # We do not want to generate bindings for some functions
+	  # We need to filter out "constructv" as well (gnome-canvas.h)
+	  $func_name = "" if ($func_name =~ /$unit_name\_constructv?$/);
+
+	  # Get_Type subprograms are handled specially, since they are
+	  # implemented with a simple pragma Import.
+	  if ($func_name =~ /$unit_name\_get_type/) {
+	    $func_name = "";
+	    $has_get_type_subprogram = 1;
+	  }
+
+	  if (($unit_name eq "" && $func_name ne "")
+	      || $func_name =~ /$prefix\_$unit_name/i) {
+	    push (@{$functions{$func_name}}, $return, @arguments);
+	  }
+	}
       }
     return %functions;
   }
 
 
-
 #########################
 ## Print the declaration for a subprogram corresponding to a gtk_..._new
 ## The second function is used to generate the equivalent Initialize function
@@ -457,7 +781,7 @@ sub print_new_declaration
     my (@arguments) = @_;
     my ($string);
     my ($indent) = "";
-    
+
     $string = "   procedure $prefix\_New";
     push (@output, $string);
     $indent = ' ' x (length ($string));
@@ -469,6 +793,7 @@ sub print_new_declaration
 sub print_initialize_declaration
 {
     my ($func_name) = shift;
+    my ($in_spec) = shift;
     my (@arguments) = @_;
     my ($string);
     my ($indent) = "";
@@ -478,7 +803,32 @@ sub print_initialize_declaration
     $indent = ' ' x (length ($string));
     &print_arguments ($indent, "void", \&convert_ada_type,
 		      3, 2, @arguments);
+    if ($in_spec) {
+      push (@output,
+	    ";\n   --  Internal initialization function.\n",
+	    "   --  See the section \"Creating your own widgets\" in "
+	    . "the documentation.\n\n");
+    }
+}
 
+#########################
+## Return the type and name of a C parameter/variable declaration
+##   Arg 1 = type definition
+##   Arg 2 = argument number
+##   Result= ($type, $name)
+#########################
+
+sub parse_c_decl {
+  my ($decl, $arg_num) = (shift, shift);
+
+  my ($type, $name) = ($decl =~ /^(.*[ *])(\w+)\s*$/);
+  if ($type eq "") {
+    $type = $_;
+    $name = "arg$arg_num";
+  }
+  $type =~ s/\s//g;
+
+  return ($type, $name);
 }
 
 #########################
@@ -500,11 +850,12 @@ sub print_arguments
     my ($for_gtk_new) = shift;
     my (@arguments) = @_;
     my ($longest) = ($return ne "void") ? 6 : 0;
+    my ($arg_num) = 1;
 
     if ($#arguments != 0
 	|| ($for_gtk_new && $#arguments > 0))   ## More than one argument ?
       {
-	$indent = (' ' x $baseindent) . "   ";
+	$indent = (' ' x $baseindent) . "  ";
 	    push (@output, "\n$indent");
       }
     elsif ($arguments[0] !~ /void/ || $for_gtk_new || $return ne "void")
@@ -512,7 +863,7 @@ sub print_arguments
 	push (@output, " ");
 	$indent .= ' ';
       }
-    
+
     if ($arguments[0] !~ /void/ || $for_gtk_new)
       {
 	push (@output, "(");
@@ -527,40 +878,20 @@ sub print_arguments
 	elsif ($for_gtk_new == 2)
 	  {
 	    push (@variables, "Widget");
-	    push (@types, "access $prefix\_$current_package\_Record");
-	  }
-	
-	foreach (@arguments)
-	  {
-	    last if (/void/);
-	    s/\s+/ /g;
-	    s/,//;
-	    s/\);//;
-	    my ($type, $name) = /^(.*[ *])(\w+)\s*$/;
-	    $type =~ s/\s//g;
-	    push (@variables, &create_ada_name ($name));
-	    $type = &{$convert} ($type);
-	    if ($type =~ /\'Class/) {
-		$type =~ s/\'Class//;
-		if ($type eq "$prefix\_$current_package") {
-		    $type = "access $type\_Record";
-		}
-		else
-		{
-		    $type = "access $type\_Record'Class";
-		}
-	    }
-	    else
-	    {
-		$type = "in $type";
-	    }
-	    push (@types, $type);
+	    push (@types, "access $prefix\_$current_package\_Record'Class");
 	  }
 
-	foreach (@variables)
-	  {
-	    $longest = length ($_) if (length ($_) > $longest);
-	  }
+	if ($arguments[0] !~ /void/) {
+	  foreach (@arguments)
+	    {
+	      my ($var, $type) = &arg_to_ada ($_, $arg_num, $convert);
+	      $arg_num ++;
+	      push (@variables, $var);
+	      $longest = length ($var) if (length ($var) > $longest);
+	      push (@types, $type);
+	    }
+	}
+
 	my ($first) = 1;
 	foreach (@variables)
 	  {
@@ -577,18 +908,16 @@ sub print_arguments
       {
 	push (@output, "\n$indent") if ($arguments[0] !~ /void/);
 	push (@output, "return ");
-	push (@output, ' ' x ($longest - 1)) if  ($arguments[0] !~ /void/);
-	if ($convert == \&convert_c_type) {
-	  if (&{$convert} ($return) eq "String") {
+	if ($convert == \&convert_c_type
+	    && &{$convert} ($return) eq "String")
+	{
 	    push (@output, "Interfaces.C.Strings.chars_ptr");
 	    $with_list {"with Interfaces.C.Strings"} ++;
-	  }
-	  else {
-	    push (@output, &{$convert} ($return));
-	  }
 	}
 	else {
-	  push (@output, &{$convert} ($return));
+	  my ($t) = &{$convert} ($return);
+	  $t =~ s/\'Class$//;
+	  push (@output, $t);
 	}
       }
   }
@@ -603,7 +932,7 @@ sub print_arguments_call
   {
     my ($indent) = shift;
     my (@arguments) = @_;
-    my ($first) = 1;
+    my ($arg_num) = 1;
 
     if ($arguments[0] !~ /void/)
       {
@@ -614,19 +943,23 @@ sub print_arguments_call
 	    s/\s+/ /g;
 	    s/,//;
 	    s/\);//;
-	    my ($type, $name) = /^(.*[ *])(\w+)\s*$/;
-	    $type =~ s/\s//g;
-	    push (@output, ",\n$indent") unless ($first);
-	    $first = 0;
+	    my ($type, $name) = &parse_c_decl ($_, $arg_num);
+	    push (@output, ",\n$indent") unless ($arg_num == 1);
+	    $arg_num++;
 	    if (&convert_c_type ($type) eq "System.Address")
 	      {
-		push (@output, "Get_Object ("
-		      .  &create_ada_name ($name) . ")");
+		if ($type eq "GtkRequisition*") {
+		  push (@output, &create_ada_name ($name)
+			. "'Address");
+		} else {
+		  push (@output, "Get_Object ("
+			.  &create_ada_name ($name) . ")");
+		}
 	      }
 	    elsif (&convert_c_type ($type) eq "String")
 	      {
 		push (@output, &create_ada_name ($name)
-		     . " & Ascii.NUL");
+		     . " & ASCII.NUL");
 	      }
 	    elsif (&convert_c_type ($type) eq "Gint"
 		  && lc ($type) ne "gint")
@@ -647,7 +980,8 @@ sub print_arguments_call
 sub print_arguments_call_for_gtk_new
 {
     my (@arguments) = @_;
-    
+    my ($arg_num) = 1;
+
     if ($arguments[0] !~ /void/)
     {
 	foreach (@arguments)
@@ -655,8 +989,8 @@ sub print_arguments_call_for_gtk_new
 	    s/\s+/ /g;
 	    s/,//;
 	    s/\);//;
-	    my ($type, $name) = /^(.*[ *])(\w+)\s*$/;
-	    $type =~ s/\s//g;
+	    my ($type, $name) = &parse_c_decl ($_, $arg_num);
+	    $arg_num++;
 	    push (@output, ",");
 	    push (@output, " " . &create_ada_name ($name));
 	  }
@@ -676,27 +1010,34 @@ sub print_declaration
     my (@arguments) = @_;
     my ($string);
     my ($indent);
-    my ($adaname) = &ada_func_name ($func_name); 
+    my ($adaname) = &ada_func_name ($func_name);
 
-    return if ($adaname eq "Get_Type");
+    if (grep (/\.\.\./, @arguments)) {
+      print STDERR "\n\n!!!No code generated for $func_name (variable number",
+	" of arguments\n";
+      push (@output,
+	    "   --  $func_name not bound: variable number of arguments\n\n");
 
-    if ($adaname =~ /New/)
-      {
-	&print_new_declaration ($func_name, @arguments);
-	push (@output, ";\n");
-	&print_initialize_declaration ($func_name, @arguments);
+    } elsif ($adaname =~ /New/) {
+      &print_new_declaration ($func_name, @arguments);
+      push (@output, ";\n\n");
+      &print_initialize_declaration ($func_name, 1, @arguments);
+      if ($has_get_type_subprogram == 1) {
+	push (@output, "   function Get_Type return Gtk.Gtk_Type;\n");
+	push (@output, "   --  Return the internal value associated with"
+	      . " this widget.\n\n");
+	$has_get_type_subprogram = -1;
       }
-    else
-      {
-	$string = ($return eq "void" ? "procedure" : "function");
-	$string .=  " $adaname";
-	push (@output, "   $string");
-	$indent = ' ' x (length ($string) + 3);
-	
-	&print_arguments ($indent, $return, \&convert_ada_type,
-			  3, 0, @arguments);
-      }
-    push (@output, ";\n");
+
+    } else {
+      $string = ($return eq "void" ? "procedure" : "function");
+      $string .=  " $adaname";
+      push (@output, "   $string");
+      $indent = ' ' x (length ($string) + 3);
+      &print_arguments ($indent, $return, \&convert_ada_type,
+			3, 0, @arguments);
+      push (@output, ";\n\n");
+    }
   }
 
 #########################
@@ -715,7 +1056,7 @@ sub print_body
     my ($indent);
     my ($adaname) = &ada_func_name ($func_name); 
 
-    return if ($adaname eq "Get_Type");
+    return if (grep (/\.\.\./, @arguments));
 
     $string = "-- $adaname --";
     push (@output, "   " . '-' x length ($string) . "\n");
@@ -736,7 +1077,7 @@ sub print_body
  	push (@output, "   " . '-' x length ($string) . "\n");
  	push (@output, "   " . $string . "\n");
  	push (@output, "   " . '-' x length ($string) . "\n\n");
- 	&print_initialize_declaration ($func_name, @arguments);
+ 	&print_initialize_declaration ($func_name, 0, @arguments);
  	$adaname = "Initialize";
       }
     else
@@ -767,26 +1108,44 @@ sub print_body
 	push (@output, $string);
 	&print_arguments_call (' ' x length ($string), @arguments);
 	push (@output, ");\n");
+	push (@output, "      Initialize_User_Data (Widget);\n");
       }
     else
       {
+	my ($terminate) = ";\n";
+
 	if (&convert_ada_type ($return) =~ /\'Class/)
 	  {
 	    my ($tmp) = &convert_ada_type ($return);
 	    $tmp =~ s/\'Class//;
-	    push (@output, "      Tmp : $tmp;\n   begin\n");
-	    $string = "      Set_Object (Tmp, Internal";
+	    push (@output, "   begin\n");
+	    if ($tmp eq "Gtk.Widget.Gtk_Widget") {
+	      $string = "      return Widget.Convert (Internal";
+	      $terminate = ");\n";
+	    } else {
+	      $string = "      return $tmp (Widget.Convert (Internal";
+	      $terminate = "));\n";
+	      }
+	    $with_list{"with Gtk.Widget; use Gtk.Widget"}++;
 	  }
 	elsif (&convert_ada_type ($return) eq "String")
 	  {
 	    push (@output, "   begin\n");
 	    $string = "      return Interfaces.C.Strings.Value (Internal";
+	    $terminate = ");\n";
 	  }
-	elsif ($return =~ /^G[dt]k/) {
+	elsif ($return eq "gboolean")
+	  {
+	    push (@output, "   begin\n");
+	    $string = "      return Boolean'Val (Internal";
+	    $terminate = ");\n";
+	  }
+	elsif ($return =~ /^(G[dt]k|Gnome)/) {
 	  push (@output, "   begin\n");
 	  $string = "      return " . &convert_ada_type ($return)
 	    . "'Val (Internal";
-	}
+	  $terminate = ");\n";
+          }
 	elsif ($return ne "void")
 	  {
 	    push (@output, "   begin\n");
@@ -798,18 +1157,9 @@ sub print_body
 	    $string = "      Internal";
 	  }
 	push (@output, $string);
-	
+
 	&print_arguments_call (' ' x length ($string), @arguments);
-	if (&convert_ada_type ($return) =~ /\'Class/) {
-	  push (@output, ");\n      return Tmp;\n");
-	}
-	elsif (&convert_ada_type ($return) eq "String"
-	      || $return =~ /^G[dt]k/) {
-	  push (@output, ");\n");
-	}
-	else {
-	  push (@output, ";\n");
-	}
+	push (@output, $terminate);
       }
     push (@output, "   end $adaname;\n\n");
   }
@@ -824,15 +1174,18 @@ sub ada_func_name
     my ($c_func_name) = shift;
     my ($type) = lc ($current_package);
 
-    $c_func_name =~ s/^(g[dt]k|ada)_?//;
+    $c_func_name =~ s/^(g[dt]k|gnome|ada)_?//;
     $c_func_name =~ s/^$type\_//;
+    $type =~ s/_//g;
+    $c_func_name =~ s/^$type\_//;  # cases like  "plot_3d" and "plot_3d_..."
 
     $type = &create_ada_name ($c_func_name);
     return "Gtk_Select" if ($type eq "Select");
+    return "Gtk_Abort" if ($type eq "Abort");
     return "$prefix\_New" if ($type =~ /New/);
     return $type;
   }
-    
+
 ###########################
 ## Generates the name for a type
 ## The current package does not appear in the generated type
@@ -843,42 +1196,67 @@ sub convert_ada_type
     my ($type) = shift;
 
     if ($type =~ /^gint([^*]*)(\*?)/) {
-      if ($2 ne "") {
-	return "out Gint$1";
-      }
-      return "Gint$1";
-    } elsif ($type =~ /^int(\*?)/) {
-      if ($1 ne "") {
-	return "out Integer";
-      }
-      return "Integer";
+      return ($2 ne "") ? "out Gint$1" : "Gint$1";
+
     } elsif ($type =~ /^guint([^*]*)(\*?)/) {
-      if ($2 ne "") {
-	return "out Guint$1";
-      }
-      return "Guint$1";
+      return ($2 ne "") ? "out Guint$1" : "Guint$1";
+
+    } elsif ($type =~ /^gdouble([^*]*)(\*?)/) {
+      return ($2 ne "") ? "out Gdouble$1" : "Gdouble$1";
+
+    } elsif ($type =~ /^double([^*]*)(\*?)/) {
+      return ($2 ne "") ? "out Double$1" : "Double$1";
+
+    } elsif ($type =~ /^int(\*?)/) {
+      return ($1 ne "") ? "out Integer" : "Integer";
+
     } elsif ($type eq "gboolean") {
       return "Boolean";
+
     } elsif ($type eq "gfloat") {
       return "Gfloat";
-    } elsif ($type =~ /(const)?(g?)char\*/) {
-      return "String";
+
+    } elsif ($type =~ /(const)?(g?)char\*(\*?)/) {
+      return ($3 eq "*") ? "Chars_Ptr_Array" : "String";
+
     } elsif ($type eq "Root_Type") {
 	return "Root_Type";
+
+    } elsif ($type eq "GtkRequisition*") {
+      $with_list {"with Gtk.Widget"} ++;
+      return "Gtk.Widget.Gtk_Requisition";
+
     } elsif ($type =~ /GSList\*/) {
-      $with_list {"with Glib.GSList"};
+      $with_list {"with Glib.GSList"} ++;
       return "GSList";
-    } elsif ($type =~ /(G[td]k)([^*]+)\*/) {
+
+    } elsif ($type =~ /Gdk([^*]+)\*/) {
+      $with_list {"with Gdk.$1"} ++;
+      return "Gdk.$1.Gdk_$1";
+
+    } elsif ($type =~ /(Gtk|Gnome)([^*]+)\*/) {
       my ($t) = $2;
       my ($prefix) = $1;
       return "Object'Class" if ($t eq "Object");
-      if ($t ne "GC") {
+
+      if ($t eq "DEntryEdit") {
+	$t = "Dentry_Edit";
+      } elsif ($t eq "MDI") {
+	$t = "MDI";
+      } elsif ($t eq "UIInfo") {
+	$t = "UI_Info";
+      } elsif ($t =~ /MDI(.+)/) {
+	$t = "MDI_$1";
+      } elsif ($t eq "Entry") {
+	$t = "GEntry";
+      } elsif ($t eq "Plot3D") {
+	$t = "Plot_3D";
+      } elsif ($t ne "GC") {
 	$t =~ s/(.)([A-Z])/$1_$2/g;
       }
-      if ("$prefix\_$t" eq "$prefix\_$current_package") {
+      if ($t =~ /$current_package/i) {
 	return "$prefix\_$t\'Class";
-      }
-      else {
+      } else {
 	$with_list {"with $prefix.$t"} ++;
 	return "$prefix.$t.$prefix\_$t\'Class";
       }
@@ -899,26 +1277,28 @@ sub convert_ada_type
 sub convert_c_type
   {
     my ($type) = shift;
-    
+
     if ($type =~ /gint([^*]*)(\*?)/) {
-      if ($2 ne "") {
-	return "out Gint$1";
-      }
-      return "Gint$1";
+      return ($2 ne "") ? "out Gint$1" : "Gint$1";
+
     } elsif ($type eq "gboolean") {
       return "Gint";
+
     } elsif ($type =~ /guint([^*]*)(\*?)/) {
-      if ($2 ne "") {
-	return "out Guint$1";
-      }
-      return "Guint$1";
+      return ($2 ne "") ? "out Guint$1" : "Guint$1";
+
+    } elsif ($type =~ /double([^*]*)(\*?)/) {
+      return ($2 ne "") ? "out Gdouble$1" : "Gdouble$1";
+
     } elsif ($type =~ /int(\*?)/) {
-      if ($1 ne "") {
-	return "out Integer";
-      }
-      return "Integer";
+      return ($1 ne "") ? "out Integer" : "Integer";
+
     } elsif ($type eq "gfloat") {
       return "Gfloat";
+
+    } elsif ($type =~ /^(Gdk[^*]*)\*/) {
+      return $1;
+
     } elsif ($type =~ /(const)?g?char\*/) {
       return "String";
     } else {
@@ -936,18 +1316,11 @@ sub print_comment
     my ($func_name) = shift;
     my ($ada) = &ada_func_name ($func_name);
     my ($string) = "";
-    
+
     $string = "   --  mapping: ";
-    if ($ada eq "Get_Type")
-      {
-	$string .= "NOT_IMPLEMENTED";
-      }
-    else
-      {
-	$string .= &ada_func_name ($func_name);
-      }
+    $string .= &ada_func_name ($func_name);
     $string .= " $hfile ";
-    if ($func_name =~ /^g[td]k/)
+    if ($func_name =~ /^(g[dt]k|gnome)/)
       {
 	if (length ($string) + length ($func_name) > 79) {
 	  $string .= "\\\n   --  mapping:      ";
